"""Package exports."""
from electric_units.electrical_energy import ElectricalEnergy
from electric_units.nem_settlement_period import NemSettlementPeriod
from electric_units.watt_sample import WattSample



"""A base SettlementPeriod class to build market specific periods from."""

from datetime import datetime, timedelta
from attr import attrs, attrib

from electric_units.utils.datetime_coercion import datetime_coercion


@attrs(frozen=True)
class BaseSettlementPeriod:
    """A Settlement Period.

    Settlement periods have a duration, exist in a timezone, and
    a means of expressing the period as an integer within a day.

    Args:
        moment: A datetime, or None to use the current time.
    """

    moment = attrib(converter=datetime_coercion, eq=False, repr=False)
    start_date = attrib(init=False, type=datetime)
    period_id = attrib(init=False, type=int)

    def __attrs_post_init__(self):
        """Post hook from attrs."""
        object.__setattr__(self, "period_id", self._period_id())
        object.__setattr__(self, "start_date", self._start_date())

    @property
    def freq_minutes(self):
        """Must be set to the region specific period duration."""
        raise NotImplementedError

    @staticmethod
    def time_zone():
        """The default timezone for this region's periods."""
        raise NotImplementedError

    @property
    def timezone(self):
        """Pull the timezone into a property of the class."""
        return self.__class__.time_zone()

    @property
    def end(self):
        """A Datetime object for the moment at the end of the period."""
        return self.start + timedelta(minutes=self.freq_minutes)

    def _start_date(self):
        """The Date the period started on."""
        return self.start.date()

    @property
    def end_date(self):
        """The Date thee period ended on."""
        return self.end.date()

    @property
    def start_hour(self):
        """An integer - the hour the period started on."""
        return self.start.hour

    @property
    def end_hour(self):
        """An integer - the hour the period ended on."""
        return self.end.hour

    def _period_id(self):
        """An integer representing the AEMO Period ID."""
        raise NotImplementedError

    @moment.default
    # pylint: disable=no-self-use
    def default_moment(self):
        """Default to UTC now."""
        return datetime.utcnow()

    @property
    def start(self):
        """A Datetime object for the moment at the start of the period.

        This object is instantiated with a moment, round this moment down
        to the start of the period.
        """
        raise NotImplementedError

    def tz_match(self, moment):
        """Localize the instantiaing moment to match this period."""
        return self.__class__.localize(moment)

    @classmethod
    def localize(cls, moment):
        """Localize a date time to match this period."""
        time_zone = cls.time_zone()

        dt_moment = datetime_coercion(moment)
        if dt_moment.tzinfo is not None:
            return dt_moment.astimezone(time_zone)
        return time_zone.localize(dt_moment)



"""A measurement or power at a given time."""

from attr import attrs, attrib

from electric_units.utils.datetime_coercion import datetime_coercion


@attrs(frozen=True)
class WattSample:
    """A sample of the power being drawn.

    A measurement, in watts, of electrical power being
    drawn at a specific moment in time.
    """

    watts = attrib(type=float)
    moment = attrib(converter=datetime_coercion)

    def settlement_period(self, period_class):
        """The period which this sample was taken within."""
        return period_class(moment=self.moment)

    @property
    def killowatts(self):
        """Convenience for more common scale of measurement - Killowatts."""
        return self.watts / 1000

    @property
    def megawatts(self):
        """Convenience for more common scale of measurement - Megawatts."""
        return self.killowatts / 1000



"""An amount of energy."""

from collections import defaultdict
from statistics import mean
from attr import attrs, attrib

from electric_units.utils.datetime_coercion import datetime_coercion


@attrs(frozen=True)
class ElectricalEnergy:
    """Energy used within a period of time."""

    kwh = attrib(type=float)
    start = attrib(converter=datetime_coercion)
    end = attrib(converter=datetime_coercion)
    samples = attrib(type=list, eq=False, repr=False, default=None)

    @classmethod
    def from_power_samples(cls, samples):
        """Create the energy object from a list of power samples.

        The energy is the weighted average of the power.

        The time between 2 samples creates a weight for the power,
        and the power being drawn between those 2 samples is the mean of those
        2 power samples.
        """
        if len(samples) < 2:
            raise TooFewSamples

        sorted_samples = sorted(samples, key=lambda sample: sample.moment)

        earliest = sorted_samples[0]
        latest = sorted_samples[-1]

        kwh = _average_kwh(sorted_samples)
        return cls(kwh=kwh,
                   start=earliest.moment, end=latest.moment,
                   samples=sorted_samples)

    @property
    def time(self):
        """Time of work, in seconds."""
        duration = self.end - self.start
        return duration.seconds

    def by_period(self, period_class):
        """Summarise the energy use in each whole settlement period."""
        if self.samples is None:
            return self._spread_energy_across_sps(period_class)

        grouped_power_samples = self._get_samples_by_period(period_class)

        energy_groups = []
        for period, p_samples in grouped_power_samples.items():
            p_samples = sorted(p_samples, key=lambda sample: sample.moment)

            first_sample = p_samples[0]
            last_sample = p_samples[-1]

            if period.tz_match(first_sample.moment) > period.start:
                start_extra = _extrapolate_constant(period.start, first_sample)
                p_samples.insert(0, start_extra)

            if period.tz_match(last_sample.moment) < period.end:
                end_extra = _extrapolate_constant(period.end, last_sample)
                p_samples.append(end_extra)

            period_energy = self.from_power_samples(p_samples)
            energy_groups.append(period_energy)

        return energy_groups

    def settlement_periods(self, period_class):
        """The periods within which this energy is used."""
        periods = []
        first_period = period_class(self.start)
        periods.append(first_period)

        while periods[-1].end < periods[-1].tz_match(self.end):
            new_period = period_class(periods[-1].end)
            periods.append(new_period)

        return periods

    def _spread_energy_across_sps(self, period_class):
        """Apply a constant mean power to all SPs within the time bound."""
        energy_groups = []
        settlement_periods = self.settlement_periods(period_class)
        mean_period_energy = self.kwh / len(settlement_periods)
        for period in settlement_periods:
            period_energy = self.__class__(
                kwh=mean_period_energy, start=period.start, end=period.end)
            energy_groups.append(period_energy)
        return energy_groups

    def _get_samples_by_period(self, period_class):
        """Return a dict, keys are the period, values are a list of samples."""
        power_groups = defaultdict(list)
        for sample in self.samples:
            sample_period = sample.settlement_period(period_class)
            tz_moment = sample_period.tz_match(sample.moment)
            tz_sample = sample.__class__(watts=sample.watts, moment=tz_moment)
            power_groups[sample_period].append(tz_sample)
        return power_groups

    def __iter__(self):
        """Iterable."""
        for key in ['kwh', 'start']:
            yield getattr(self, key)


def _extrapolate_constant(to_time, from_sample):
    """Extrapolate a power rating to an earlier or later time.

    This models a constant power rating before (or after)
    the `from_sample` input to the `to_time`.

    It returns a power sample, of the same class as the `from_sample`
    but at the time specified in `to_time`.
    """
    sample_class = from_sample.__class__
    watts = from_sample.watts
    extrapolated_sample = sample_class(watts=watts, moment=to_time)
    return extrapolated_sample


def _average_kwh(power_samples):
    """Average khw of a group of samples."""
    if len(power_samples) < 2:
        raise TooFewSamples

    kwh = 0
    for sample_1, sample_2 in zip(power_samples[:-1], power_samples[1:]):
        watts = mean([sample_1.watts, sample_2.watts])
        duration = (sample_2.moment - sample_1.moment).seconds
        kwh += (watts / 1000) * (duration / 3600)

    return kwh


class TooFewSamples(IndexError):
    """There are not enough power sampels to represent an energy."""



"""Functions to evaluate the maximum extractable flex."""
from os import path
from setuptools import setup, find_packages

# read the contents of your README file
REPO_DIR = path.abspath(path.dirname(__file__))
with open(path.join(REPO_DIR, 'README.md'), encoding='utf-8') as f:
    LONG_DESCRIPTION = f.read()

setup(
    name='electric_units',
    version='1.0.2',
    description='For handling units of measurement in electricity grid data.',
    long_description=LONG_DESCRIPTION,
    long_description_content_type='text/markdown',
    url='http://github.com/ample-tech/electric-units',
    author='ample sustainable brands',
    author_email='administrator@ample.tech',
    zip_safe=False,
    license='MIT',
    packages=find_packages(exclude=['tests', 'tests.*']),
    include_package_data=True,
    keywords=['energy', 'grid', 'data', 'data transformation'],
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'Topic :: Scientific/Engineering :: Information Analysis',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3',
    ],
    setup_requires=['pytest-runner', 'pytest-pylint'],
    tests_require=['pytest', 'pylint'],
    install_requires=[
        'attrs',
        'numpy',
        'pandas>0.25',
        'python-dateutil',
        'pytz',
        'requests',
    ]
)



"""Convert from a string to a datetime."""

from dateutil.parser import parse


def datetime_coercion(moment):
    """Force the return of a datetime object."""
    if isinstance(moment, str):
        return parse(moment)
    return moment



"""Pandas DType and Array for NemSettlementPeriod."""
import sys
import numpy as np

from pandas.api.extensions import register_extension_dtype, take
from pandas.api.types import is_list_like, is_scalar
from pandas.core.arrays import ExtensionArray
from pandas.core.dtypes.base import ExtensionDtype


from electric_units.nem_settlement_period import NemSettlementPeriod


@register_extension_dtype
class NemSettlementPeriodDtype(ExtensionDtype):
    """A custom data type, to be paired with an ExtensionArray."""

    type = NemSettlementPeriod
    name = "nem_settlement_period"

    @classmethod
    def construct_array_type(cls):
        """Return the array type associated with this dtype."""
        return NemSettlementPeriodArray


class NemSettlementPeriodArray(ExtensionArray):
    """Abstract base class for custom 1-D array types."""

    def __init__(self, values, dtype=NemSettlementPeriodDtype, copy=False):
        """Instantiate the array."""
        if copy:
            values = values.copy()
        if not isinstance(values[0], NemSettlementPeriod):
            values = [NemSettlementPeriod(val) for val in values]
        self.data = np.asarray(values, dtype=object)
        self._dtype = dtype()

    def __setitem__(self, key, value):
        """Set one or more values inplace."""
        if is_list_like(value):
            if is_scalar(key):
                raise ValueError("setting an array element with a sequence.")
            value = [NemSettlementPeriod(val) for val in value]
        else:
            value = NemSettlementPeriod(value)
        self.data[key] = value

    @classmethod
    def _from_sequence(cls, scalars, dtype=None, copy=False):
        """Construct a new ExtensionArray from a sequence of scalars."""
        return cls(scalars)

    @classmethod
    def _from_factorized(cls, values, original):
        """Reconstruct an ExtensionArray after factorization."""
        values = [val.decode('utf-8') for val in values]
        return cls(values)

    def _values_for_factorize(self):
        """Factorize to the ISO8601, with TZ, of the start datetime."""
        starts = np.fromiter((s.start.isoformat() for s in self), '|S25')
        return np.array(starts, dtype=object, copy=True), np.nan

    def __getitem__(self, item):
        """Select a subset of self."""
        return self.data[item]

    def __len__(self):
        """Length of this array."""
        return len(self.data)

    @property
    def nbytes(self):
        """The byte size of the data."""
        return len(self) * sys.getsizeof(self[0])

    @property
    def dtype(self):
        """An instance of 'ExtensionDtype'."""
        return self._dtype

    def isna(self):
        """A 1-D array indicating if each value is missing."""
        return np.array([False for x in self.data], dtype=bool)

    def take(self, indices, allow_fill=False, fill_value=None):
        """Take elements from an array.

        Relies on the take method defined in pandas:
        https://github.com/pandas-dev/pandas/blob/e246c3b05924ac1fe083565a765ce847fcad3d91/pandas/core/algorithms.py#L1483
        """
        data = self.data
        if allow_fill and fill_value is None:
            fill_value = self.dtype.na_value

        result = take(
            data, indices, fill_value=fill_value, allow_fill=allow_fill)
        return self._from_sequence(result)

    def copy(self):
        """Return a copy of the array."""
        return type(self)(self.data.copy())

    @classmethod
    def _concat_same_type(cls, to_concat):
        """Concatenate multiple arrays."""
        return cls(np.concatenate([x.data for x in to_concat]))



"""A SettlementPeriod in the NEM region."""
from datetime import timedelta
from pytz import timezone
from attr import attrs

from electric_units.base_settlement_period import BaseSettlementPeriod


@attrs(frozen=True)
class NemSettlementPeriod(BaseSettlementPeriod):
    """A NEM Settlement Period.

    NEM has 30 minute periods. Instantiate one with a moment in time.
    This will assume this moment is in AEST timezone, or, if a Timezone is
    present in the instantiating moment, then it will be converted to AEST.

    Args:
        moment: A datetime, or None to use the current time.
    """

    @property
    def freq_minutes(self):
        """The NEM has 30 minute periods."""
        return 30

    @staticmethod
    def time_zone():
        """Australian Eastern Standard Time."""
        return timezone('Etc/GMT-10')

    @property
    def start(self):
        """Rewind to the start of the half hour."""
        moment = self.tz_match(self.moment)

        mins = moment.minute
        if mins < 30:
            round_by = mins
        else:
            round_by = (30 - mins) * -1

        start = moment - timedelta(minutes=round_by)
        start = start.replace(second=0, microsecond=0)
        return start

    def _period_id(self):
        """An integer representing the AEMO Period ID within the day."""
        start_time = self.start.time()
        hour = start_time.hour
        period_id = (hour * 2) + 1

        minute = start_time.minute
        if minute >= 30:
            period_id = period_id + 1

        return period_id



"""Convert a series."""


def to_settlement_period(extension_array, series_of_strings):
    """Convert from a series from strings to settlement period objects."""
    return extension_array(series_of_strings)
